import types

from django.db import models
from django.db.models.manager import BaseManager


class _SerializationWrapper:
    """Method wrapper to make a picklable object."""

    def __init__(self, serializer, obj):
        self.serializer = serializer
        self.obj = obj

    def __call__(self):
        return self.serializer._serialize_object(self.obj)


class BaseSerializer:
    select_related = None
    prefetch_related = None
    extra = None

    def __init__(self, *, select_related=None, prefetch_related=None, extra=None):
        if select_related:
            self.select_related = select_related
        if prefetch_related:
            self.prefetch_related = prefetch_related

        if extra:
            self.extra = {key: _resolve_serializer(self.extra[key]) for key in extra}
        else:
            self.extra = {}

    def _prepare_queryset(self, qs):
        if self.select_related:
            if callable(self.select_related):
                qs = qs.select_related(*self.select_related())
            else:
                qs = qs.select_related(*self.select_related)

        if self.prefetch_related:
            if callable(self.prefetch_related):
                qs = qs.prefetch_related(*self.prefetch_related())
            else:
                qs = qs.prefetch_related(*self.prefetch_related)

        qs = self.prepare_queryset(qs)

        for extra in self.extra.values():
            qs = extra._prepare_queryset(qs)

        return qs

    def prepare_queryset(self, qs):
        """
        Custom change the queryset. It is possible to implement `select_related`
        and `prefetch_related` attributes with it, but they work nice together.
        """
        return qs

    def _prepare_objects(self, objs):
        self.prepare_objects(objs)

        for extra in self.extra.values():
            extra._prepare_objects(objs)

        for obj in objs:
            obj.serialize = _SerializationWrapper(self, obj)

    def prepare_objects(self, objs):
        """
        Prepare objects after they are loaded to memory.

        It is a hook to add data in bulk to loaded objects, like fetching info
        from cache and attaching to them.
        """
        pass

    def _serialize_object(self, obj):
        serialized = self.serialize_object(obj)
        for extra in self.extra.values():
            serialized.update(extra._serialize_object(obj))
        return serialized

    def serialize_object(self, obj):
        """
        Required implementation. It converts the Django model to a serializable
        dict.

        Avoid slow calls here because it will cause N+1 issues.
        """
        raise NotImplementedError

    def serialize(self, objs):
        yield from map(self._serialize_object, objs)


class SerializableQuerySet(models.QuerySet):
    @property
    def serializer(self):
        return getattr(self, '_serializer', None)

    def to_serialize(self, serializer=None):
        self._serializer = _resolve_serializer(serializer)
        return self._serializer._prepare_queryset(self)

    def _fetch_all(self):
        super()._fetch_all()
        self._serializer._prepare_objects(self._result_cache)

    def _clone(self):
        c = super()._clone()
        c._serializer = self.serializer
        return c


class SerializableManager(BaseManager.from_queryset(SerializableQuerySet)):
    def __init__(self, *, select_related=None, prefetch_related=None,
                 default_serializer=BaseSerializer):
        super().__init__()
        self.default_serializer = default_serializer(
            select_related=select_related,
            prefetch_related=prefetch_related,
        )

    def to_serialize(self, serializer=None):
        if serializer is None:
            serializer = self.default_serializer
        return self.get_queryset().to_serialize(serializer)


def serialize(objs):
    try:
        obj = objs[0]
    except IndexError:
        return []
    serializer = obj.serialize.serializer
    return serializer.serialize(objs)


class _FuncSerializer(BaseSerializer):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def serialize_object(self, obj):
        return self.func(obj)


def _resolve_serializer(serializer):
    if isinstance(serializer, types.FunctionType):
        serializer = _FuncSerializer(serializer)
    elif not isinstance(serializer, BaseSerializer):
        serializer = serializer()
    return serializer



from django.db import models

from django_qserializer.serialization import SerializableManager


class Company(models.Model):
    name = models.CharField(max_length=64)


class Bus(models.Model):
    plate = models.CharField(max_length=10)
    company = models.ForeignKey(Company, on_delete=models.SET_NULL)
    objects = SerializableManager()


class Travel(models.Model):
    bus = models.ForeignKey(Bus, related_name='travels', on_delete=models.SET_NULL)



import pytest

from django_qserializer import BaseSerializer, serialize
from django_qserializer.tests.testapp.models import Bus, Company, Travel


@pytest.fixture
def bus_fixture(db):
    company = Company.objects.create(name='Hurricane Cart')
    return Bus.objects.create(company=company, plate='BUSER')


@pytest.fixture
def travel_fixture(db, bus_fixture):
    return Travel.objects.create(bus=bus_fixture)


def test_magic_serialize_method(bus_fixture, django_assert_num_queries):
    class S(BaseSerializer):
        select_related = ['company']

        def serialize_object(self, bus):
            return {
                'company': bus.company.name,
            }

    bus = Bus.objects.to_serialize(S).first()
    with django_assert_num_queries(0):
        assert {'company': 'Hurricane Cart'} == bus.serialize()


def test_global_serialize(bus_fixture, django_assert_num_queries):
    class S(BaseSerializer):
        select_related = ['company']

        def serialize_object(self, bus):
            return {
                'company': bus.company.name,
            }

    bus = Bus.objects.to_serialize(S).first()
    with django_assert_num_queries(0):
        assert [{'company': 'Hurricane Cart'}] == list(serialize([bus]))


def test_global_serialize_empty():
    assert [] == serialize([])


def test_serialize_object_not_implemented(bus_fixture):
    bus = Bus.objects.to_serialize().first()
    with pytest.raises(NotImplementedError):
        bus.serialize()


def test_extras(bus_fixture, django_assert_num_queries):
    class Attr(BaseSerializer):
        select_related = ['company']

        def serialize_object(self, obj):
            return {
                'myattr': obj.company.name
            }

    def func(obj):
        return {
            'seats': 32,
        }

    class S(BaseSerializer):
        extra = {
            'myattr': Attr,
            'func': func,
        }

        def serialize_object(self, obj):
            return {
                'plate': obj.plate,
            }

    serializer = S(extra=['myattr', 'func'])

    with django_assert_num_queries(1):
        bus = Bus.objects.to_serialize(serializer).first()

    expected = {
        'plate': 'BUSER',
        'myattr': 'Hurricane Cart',
        'seats': 32,
    }

    with django_assert_num_queries(0):
        assert expected == bus.serialize()

    with django_assert_num_queries(0):
        assert expected == next(serialize([bus]))


def test_extras_recursive(bus_fixture, django_assert_num_queries):
    def city(obj):
        return {
            'city': 'SJK',
        }

    class Attr(BaseSerializer):
        extra = {
            'city': city,
        }
        select_related = ['company']

        def serialize_object(self, obj):
            return {
                'myattr': obj.company.name
            }

    class S(BaseSerializer):
        extra = {
            'myattr': Attr(extra=['city']),
        }

        def serialize_object(self, obj):
            return {
                'plate': obj.plate,
            }

    serializer = S(extra=['myattr'])

    with django_assert_num_queries(1):
        bus = Bus.objects.to_serialize(serializer).first()

    expected = {
        'plate': 'BUSER',
        'myattr': 'Hurricane Cart',
        'city': 'SJK',
    }

    with django_assert_num_queries(0):
        assert expected == bus.serialize()


def test_prepare_objects_after_prefetch(travel_fixture):
    """
    Regression test. Prior implementation ran prepare_objects before prefetchs.
    """

    class S(BaseSerializer):
        prefetch_related = ['travels']

        def prepare_objects(self, objs):
            for obj in objs:
                assert obj._prefetched_objects_cache

        def serialize_object(self, obj):
            return {
                'plate': obj.plate,
            }

    Bus.objects.to_serialize(S).first()



from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


SECRET_KEY = 'wewls(#^h)7)(8x6l)8-d@$!hv909h48f3y3z*04_m3_5_*n^$'

DEBUG = True

INSTALLED_APPS = [
    'django_qserializer.tests.testapp',
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': BASE_DIR / 'db.sqlite3',
    }
}



from .serialization import (
    BaseSerializer,
    SerializableManager,
    SerializableQuerySet,
    serialize,
)

BaseSerializer = BaseSerializer
SerializableManager = SerializableManager
SerializableQuerySet = SerializableQuerySet
serialize = serialize



from setuptools import find_packages, setup

setup(
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'django',
    ],
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=[
        'pytest',
        'pytest-cov',
        'pytest-django',
        'pytest-flake8',
    ],
)
